/*
 ============================================================================
 Name        : jnm_decodingplayer.c
 Author      : Klaus Jaensch
 Version     :
 Copyright   : Your copyright notice
 Description : Java native media decoding player
 ============================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>

#include <gst/gstutils.h>
#include <gst/gst.h>
#include <gst/video/gstvideosink.h>
#include <gst/video/videooverlay.h>
#include <gst/audio/gstaudiosink.h>
#include <X11/Xlib.h>
#include <jawt.h>
#include <jawt_md.h>

#include "jnm_audiosink.h"
#include "jnm_videosink.h"

#include "jnm_decodingplayer.h"

//GST_BOILERPLATE(JnmDecodingPlayer, jnm_decodingplayer, GstObject, GST_TYPE_OBJECT);
G_DEFINE_TYPE_WITH_PRIVATE(JnmDecodingPlayer, jnm_decodingplayer,GST_TYPE_OBJECT);
extern jweak exceptionClassRef;


static void callJavaVoidMethod(JnmDecodingPlayer *mdp, char * methodName) {
	JNIEnv *env;
	JavaVM *jvm = mdp->javaVM;
	if (jvm != NULL) {
		jint res = (*jvm)->AttachCurrentThread(jvm, (void **) &env, NULL);
		if (res == 0) {
			jclass cls = (*env)->GetObjectClass(env, mdp->javaObjRef);
			if (cls == NULL) {
				fprintf(stderr,"Could not find class\n");
				fflush(stderr);
				return;
			}
			jmethodID mid = (*env)->GetMethodID(env, cls, methodName, "()V");
			if (mid == 0) {
				fprintf(stderr,"Could not find method ID\n");
				fflush(stderr);
				return;
			}
			(*env)->CallVoidMethod(env, mdp->javaObjRef, mid);
			(*jvm)->DetachCurrentThread(jvm);

		}
	}else{
		fprintf(stderr,"NO JVM reference\n");
		fflush(stderr);
	}
}

static void callJavaErrorMethod(JnmDecodingPlayer *mdp, char * errorMsg) {
	JNIEnv *env;
	JavaVM *jvm = mdp->javaVM;
	if (jvm != NULL) {
		jint res = (*jvm)->AttachCurrentThread(jvm, (void **) &env, NULL);
		if (res == 0) {
			jclass cls = (*env)->GetObjectClass(env, mdp->javaObjRef);
			if (cls == NULL) {
				fprintf(stderr,"Could not find class\n");
				fflush(stderr);
				return;
			}
			jmethodID mid = (*env)->GetMethodID(env, cls, "error",
					"(Ljava/lang/String;)V");
			if (mid == 0) {
				fprintf(stderr,"Could not find method ID\n");
				fflush(stderr);
				return;
			}
			jstring errorMsgJ = (*env)->NewStringUTF(env, errorMsg);
			(*env)->CallVoidMethod(env, mdp->javaObjRef, mid, errorMsgJ);
			(*jvm)->DetachCurrentThread(jvm);
		}
	}
}


static void callJavaDurationAvailable(JnmDecodingPlayer *mdp, jlong duration) {
	JNIEnv *env;
	JavaVM *jvm = mdp->javaVM;
	if (jvm != NULL) {
		jint res = (*jvm)->AttachCurrentThread(jvm, (void **) &env, NULL);
		if (res == 0) {
			jclass cls = (*env)->GetObjectClass(env, mdp->javaObjRef);
			if (cls == NULL) {
				fprintf(stderr,"Could not find class\n");
				fflush(stderr);
				return;
			}
			jmethodID mid = (*env)->GetMethodID(env, cls, "durationAvailable", "(J)V");
			if (mid == 0) {
				fprintf(stderr,"Could not find method ID\n");
				fflush(stderr);
				return;
			}
			(*env)->CallVoidMethod(env, mdp->javaObjRef, mid,duration);
			(*jvm)->DetachCurrentThread(jvm);

		}
	}
}

static gboolean jnm_decodingplayer_bus_callback(GstBus *bus,
		GstMessage *message, gpointer data) {
	// g_print("Got %s message\n", GST_MESSAGE_TYPE_NAME (message));
	JnmDecodingPlayer *mdp = (JnmDecodingPlayer *) data;

	if (gst_is_video_overlay_prepare_window_handle_message (message)){
		gst_video_overlay_set_window_handle (GST_VIDEO_OVERLAY (GST_MESSAGE_SRC (message)),mdp->drawable);
	}
	switch (GST_MESSAGE_TYPE (message)) {
	case GST_MESSAGE_ERROR: {
		GError *err;
		gchar *debug;

		gst_message_parse_error(message, &err, &debug);
		g_printerr("Error: %s %s\n", err->message,debug);
		callJavaErrorMethod(mdp, err->message);
		g_error_free(err);
		g_free(debug);
		break;
	}

	case GST_MESSAGE_STATE_CHANGED: {
		if (message->src != (GstObject *) mdp->play) {
			// ignore messages from other components
			return TRUE;
		}
		GstState oldState;
		GstState newState;
		GstState pendingState;
		gst_message_parse_state_changed(message, &oldState, &newState,
				&pendingState);
//
//		g_print("Element %s changed state from %s to %s (Pending %s).\n",
//		GST_OBJECT_NAME (message->src),
//				gst_element_state_get_name(oldState),
//				gst_element_state_get_name(newState),
//				gst_element_state_get_name(pendingState));

		if (oldState == GST_STATE_NULL && newState == GST_STATE_READY) {
			// Java callback opened
			callJavaVoidMethod(mdp, "opened");
		} else if (newState == GST_STATE_PLAYING) {
			// Java callback started
			callJavaVoidMethod(mdp, "started");
		} else if (newState == GST_STATE_NULL) {
			// play bin closed
			// quit the main loop
			g_main_loop_quit(mdp->loop);

			// stop watching bus events (return FALSE)
			return FALSE;
		}
		break;
	}
	case GST_MESSAGE_ASYNC_DONE: {
		jlong duration=jnm_decodingplayer_getDuration(mdp);
		if(duration>0){
			//g_print("Duration %lld .\n",duration);
			callJavaDurationAvailable(mdp,duration);
		}
		break;
	}
	case GST_MESSAGE_EOS: {
		callJavaVoidMethod(mdp, "endOfStream");
		break;
	}
	default:
		/* unhandled message */
		break;
	}
	// Still want to receive messages (return TRUE)
	return TRUE;
}

static GstStateChangeReturn jnm_decodingplayer_open(JnmDecodingPlayer *filter) {

	g_object_set(G_OBJECT (filter->play), "uri", filter->uri, NULL);

	GstPipeline *pipeline = (GstPipeline*) filter->play;
	// we want to see the close state change READY-> NULL
	gst_pipeline_set_auto_flush_bus(pipeline, FALSE);

	GstBus *bus = gst_pipeline_get_bus(GST_PIPELINE (filter->play));
	gst_bus_add_watch(bus, jnm_decodingplayer_bus_callback, filter);

	gst_object_unref(bus);

	return gst_element_set_state((GstElement *) filter->play, GST_STATE_READY);

}

/* Init functions */
static void jnm_decodingplayer_class_init(JnmDecodingPlayerClass *klass) {
	//GstObjectClass *gobject_class= (GstObjectClass *) klass;
	// TODO does not compile anymore
	//
	//parent_class = g_type_class_peek_parent(klass);
}

/*
static void jnm_decodingplayer_base_init(gpointer g_class) {

}
*/

static void jnm_decodingplayer_init(JnmDecodingPlayer *filter) {
	filter->loop = g_main_loop_new(NULL, FALSE);
	filter->play = (GstBin *) gst_element_factory_make("playbin", "play");
}

static void jnm_decodingplayer_start(JnmDecodingPlayer *filter) {
	gst_element_set_state((GstElement *) filter->play, GST_STATE_PLAYING);

}
static void jnm_decodingplayer_stop(JnmDecodingPlayer *filter) {
	gst_element_set_state((GstElement *) filter->play, GST_STATE_PAUSED);
}

static jboolean jnm_decodingplayer_seek(JnmDecodingPlayer *filter,guint64 pos) {
	GstBin *play=filter->play;
	GstElement *playE=(GstElement*)play;
	return gst_element_seek_simple(playE, GST_FORMAT_TIME, GST_SEEK_FLAG_FLUSH,pos);
}

static void jnm_decodingplayer_frameStep(JnmDecodingPlayer *filter,guint64 amount) {
	GstBin *play=filter->play;
	gdouble rate=1.0;
	GstVideoSink *videoSink;
	g_object_get (play, "video-sink", &videoSink, NULL);
	GstElement *vsElem=(GstElement *)videoSink;
	if(amount<0){
		 GstEvent *seEv=gst_event_new_seek(rate,GST_FORMAT_BUFFERS,0,GST_SEEK_TYPE_NONE,0,GST_SEEK_TYPE_NONE,0);
		 GstElement *playE=(GstElement*)play;
		 gst_element_send_event (playE,seEv);
	}
	GstEvent *stepev=gst_event_new_step(GST_FORMAT_BUFFERS, amount, 1.0, TRUE, FALSE);

	gst_element_send_event (vsElem,stepev);
}

static void jnm_decodingplayer_close(JnmDecodingPlayer *mdp) {
	gst_element_set_state((GstElement *) mdp->play, GST_STATE_NULL);
}
static void jnm_decodingplayer_release(JnmDecodingPlayer *filter) {
	// TODO do we have to unref the video/audio sinks too?

	free(filter->uri);
	gst_object_unref(GST_OBJECT (filter->play));
	gst_object_unref(GST_OBJECT (filter));

}
static void jnm_decodingplayer_rewind(JnmDecodingPlayer *mdp) {
	GstElement *playElement = (GstElement *) mdp->play;
	gst_element_seek_simple(playElement, GST_FORMAT_DEFAULT,
			GST_SEEK_FLAG_FLUSH | GST_SEEK_FLAG_KEY_UNIT, 0);
}


jlong jnm_decodingplayer_getDuration(JnmDecodingPlayer *mdp) {
	GstFormat fmt = GST_FORMAT_TIME;
	gint64 len;
	GstElement *playElement = (GstElement *) mdp->play;

	if(gst_element_query_duration(playElement, fmt, &len)) {
		return len;
	}else{
		return -1;
	}
}

static jlong jnm_decodingplayer_getPosition(JnmDecodingPlayer *mdp) {
	GstFormat fmt = GST_FORMAT_TIME;
	gint64 pos;
	GstElement *playElement = (GstElement *) mdp->play;
	if (gst_element_query_position(playElement, fmt, &pos)){
		return pos;
	}else{
		return -1;
	}
}
gboolean jnm_decodingplayer_set_caps(JnmDecodingPlayer *mdp, GstCaps *caps) {
	gint width, height;

	if (gst_caps_is_fixed(caps)) {

		const GstStructure *str = gst_caps_get_structure(caps, 0);
		if (gst_structure_get_int(str, "width", &width)
				&& gst_structure_get_int(str, "height", &height)) {

			/*
			 g_print ("The video size of this set of capabilities is %dx%d\n",
			 width, height);
			 */
		}

		//g_print("Caps: %s\n",gst_structure_to_string(str));

		gint parNumerator = 1;
		gint parDenominator = 1;
		if (gst_structure_has_field(str, "pixel-aspect-ratio")) {
			gst_structure_get_fraction(str, "pixel-aspect-ratio",
					&parNumerator, &parDenominator);
		}
		JNIEnv *env;
		JavaVM *jvm = mdp->javaVM;

		jint res = (*jvm)->AttachCurrentThread(jvm, (void **) &env, NULL);
		if (res == 0) {

			jclass cls = (*env)->GetObjectClass(env, mdp->javaObjRef);
			if (cls == NULL) {
				g_printerr("Could not find class\n");
				return FALSE;
			}
			jmethodID mid = (*env)->GetMethodID(env, cls, "setVideoSize",
					"(IIII)[B");
			if (mid == 0) {
				g_printerr("Could not find method ID\n");
				return FALSE;
			}

			jobject imgBufArr = (*env)->CallObjectMethod(env, mdp->javaObjRef,
					mid, width, height, parNumerator, parDenominator);
			mdp->imgBufferArr = (*env)->NewWeakGlobalRef(env, imgBufArr);
			(*jvm)->DetachCurrentThread(jvm);
			return TRUE;
		}
	}
	return FALSE;
}

GstFlowReturn jnm_decodingplayer_show_frame(JnmDecodingPlayer *mdp,
		GstBuffer *imgBuf) {
	pthread_t currNativeThreadId=pthread_self();
	JNIEnv *env;
	JavaVM *jvm = mdp->javaVM;
	if(mdp->videoNativeThreadID!=currNativeThreadId){
		jint res = (*jvm)->AttachCurrentThread(jvm, (void **) &env, NULL);
		if (res == 0) {
			mdp->videoNativeThreadID=currNativeThreadId;
			mdp->env=env;
		}else{
			g_printerr("Could not attach thread to JVM!");
			env=NULL;
			mdp->env=NULL;
		}
	}else{
		env=mdp->env;
	}
	if(env!=NULL){

		GstMemory*imgMem=gst_buffer_get_all_memory (imgBuf);
		GstMapInfo imgMapInfo;
		if(gst_memory_map (imgMem,&imgMapInfo,GST_MAP_READ)){
			jsize dataLength = imgMapInfo.size;
			const jbyte *data = (jbyte *) imgMapInfo.data;
			int arrLen=(*env)->GetArrayLength(env,mdp->imgBufferArr);
			if(data!=NULL && arrLen==dataLength){
				(*env)->SetByteArrayRegion(env, mdp->imgBufferArr, 0, dataLength, data);
				(*env)->CallVoidMethod(env, mdp->javaObjRef,(jmethodID) mdp->jRenderMethodId);
			}
		}
		return GST_FLOW_OK;
	}
	return GST_FLOW_ERROR;
}


JNIEXPORT jobject JNICALL Java_de_klausj_video_impl_gstreamer_GstMediaDecodingPlayer_init(JNIEnv *env, jobject jobj) {
	jobject nativePointer = NULL;
	JavaVM *javaVM;
	jint res = (*env)->GetJavaVM(env, &javaVM);
	if (res == 0) {

		JnmDecodingPlayer *jnmDecPlayer = g_object_new(
				jnm_decodingplayer_get_type(), NULL);

		jnmDecPlayer->javaVM = javaVM;
		jnmDecPlayer->javaObjRef = (*env)->NewWeakGlobalRef(env, jobj);

		nativePointer = (*env)->NewDirectByteBuffer(env, jnmDecPlayer,
				sizeof(JnmDecodingPlayer));
		jclass cls = (*env)->GetObjectClass(env, jnmDecPlayer->javaObjRef);
		if (cls == NULL) {
			g_printerr("Could not find class\n");
			return NULL;
		}
		jnmDecPlayer->jClassRef=(*env)->NewGlobalRef(env,cls);
		jnmDecPlayer->jRenderMethodId = (*env)->GetMethodID(env, cls, "render", "()V");
		if (jnmDecPlayer->jRenderMethodId == 0) {
			g_printerr("Could not find method ID\n");
			return NULL;
		}
	}
	return nativePointer;
}

JNIEXPORT jlong JNICALL Java_de_klausj_video_impl_gstreamer_GstMediaDecodingPlayer_nativeWindowHandleX11Drawable(
		JNIEnv *env, jobject jobj,jobject xWindowCanvas) {

	// we need the X11 window ID (XID) of the Java component
	// the code follows examples in the JDK include files jawt.h and linux/jawt_md.h
	JAWT jawt;
	jawt.version = JAWT_VERSION_1_3;
	jboolean res=JAWT_GetAWT(env,&jawt);
	if(!res){

		(*env)->ThrowNew(env,(jclass) exceptionClassRef,"Could not get AWT!");
	}else{
		JAWT_DrawingSurface* jawtDs=jawt.GetDrawingSurface(env,xWindowCanvas);
		if(jawtDs==NULL){
			(*env)->ThrowNew(env,(jclass) exceptionClassRef,"Could not get drawing surface");

		}else{
			jawtDs->Lock(jawtDs);
			JAWT_DrawingSurfaceInfo* dsInfo=jawtDs->GetDrawingSurfaceInfo(jawtDs);
			if(dsInfo==NULL){
				(*env)->ThrowNew(env,(jclass) exceptionClassRef,"Could not get drawing surface info");
			}else{
				JAWT_X11DrawingSurfaceInfo *x11DsInfo=(JAWT_X11DrawingSurfaceInfo*)(dsInfo->platformInfo);
				Drawable drawable=x11DsInfo->drawable;
				jawtDs->FreeDrawingSurfaceInfo(dsInfo);
				jawtDs->Unlock(jawtDs);
				jawt.FreeDrawingSurface(jawtDs);

				// Drawable is an XID
				// printf("Window ID: %ld\n",drawable);
				return drawable;
			}
		}
	}
	return 0;
}

JNIEXPORT void JNICALL Java_de_klausj_video_impl_gstreamer_GstMediaDecodingPlayer_configureVideoSink(
		JNIEnv *env, jobject jobj,jobject nativePointer,jlong x11Drawable) {
	JnmDecodingPlayer *jnmDecPlayer = (*env)->GetDirectBufferAddress(env,
			nativePointer);

	if(x11Drawable==0){
		JnmVideoSink *jvs = g_object_new(JNM_TYPE_VIDEOSINK, NULL);
		jvs->mediaDecodingPlayer = jnmDecPlayer;
		g_object_set(G_OBJECT (jnmDecPlayer->play), "video-sink", jvs, NULL);
	}else{
		GstVideoSink * xvImgSink=(GstVideoSink *) gst_element_factory_make("xvimagesink", NULL);
		g_object_set(G_OBJECT(xvImgSink),"force-aspect-ratio",TRUE,NULL);
		g_object_set(G_OBJECT (jnmDecPlayer->play), "video-sink", xvImgSink, NULL);
		jnmDecPlayer->xoverlay=(GstVideoOverlay *)xvImgSink;
	}
}

JNIEXPORT void JNICALL Java_de_klausj_video_impl_gstreamer_GstMediaDecodingPlayer_setVideoSinkWindow(
				JNIEnv *env, jobject jobj,jobject nativePointer,jlong x11Drawable) {
			JnmDecodingPlayer *jnmDecPlayer = (*env)->GetDirectBufferAddress(env,
					nativePointer);
			if(x11Drawable!=0 && jnmDecPlayer->xoverlay !=NULL){
				gst_video_overlay_set_window_handle (jnmDecPlayer->xoverlay, x11Drawable);
				gst_video_overlay_expose (jnmDecPlayer->xoverlay);
			}


}

JNIEXPORT void JNICALL Java_de_klausj_video_impl_gstreamer_GstMediaDecodingPlayer_setAudioSink(
		JNIEnv *env, jobject jobj,jobject nativePointer,jobject audioSink) {
	JnmDecodingPlayer *jnmDecPlayer = (*env)->GetDirectBufferAddress(env,
			nativePointer);
	JnmAudioSink *jnmAudioSink = (*env)->GetDirectBufferAddress(env,audioSink);
	jnmDecPlayer->audioSink=jnmAudioSink;
	g_object_set(G_OBJECT (jnmDecPlayer->play), "audio-sink", jnmDecPlayer->audioSink, NULL);

}

JNIEXPORT void JNICALL Java_de_klausj_video_impl_gstreamer_GstMediaDecodingPlayer_gMainLoopRun(
		JNIEnv *env, jobject jobj, jobject nativePointer) {
	JnmDecodingPlayer *jnmDecPlayer = (*env)->GetDirectBufferAddress(env,
			nativePointer);
	g_main_loop_run(jnmDecPlayer->loop);
}

JNIEXPORT void JNICALL Java_de_klausj_video_impl_gstreamer_GstMediaDecodingPlayer_open(
		JNIEnv *env, jobject jobj, jobject nativePointer, jstring mediaFilePath) {
	JnmDecodingPlayer *jnmDecPlayer = (*env)->GetDirectBufferAddress(env,
			nativePointer);
	jboolean copy;
	const char *uri = (*env)->GetStringUTFChars(env, mediaFilePath, &copy);
	if (uri == NULL) {
		return;
	}
	size_t uriLen = strlen(uri);
	jnmDecPlayer->uri = malloc(uriLen + 1);
	strncpy(jnmDecPlayer->uri, uri, uriLen + 1);
	(*env)->ReleaseStringUTFChars(env, mediaFilePath, uri);

	GstStateChangeReturn scr = jnm_decodingplayer_open(jnmDecPlayer);
	if (scr == GST_STATE_CHANGE_FAILURE) {
		(*env)->ThrowNew(env, (jclass) exceptionClassRef,
				"Could not prepare playback engine");
	}
}

JNIEXPORT void JNICALL Java_de_klausj_video_impl_gstreamer_GstMediaDecodingPlayer_start(
		JNIEnv *env, jobject jobj, jobject nativePointer) {
	JnmDecodingPlayer *jnmDecPlayer = (*env)->GetDirectBufferAddress(env,
			nativePointer);
	jnm_decodingplayer_start(jnmDecPlayer);
}
JNIEXPORT void JNICALL Java_de_klausj_video_impl_gstreamer_GstMediaDecodingPlayer_stop(
		JNIEnv *env, jobject jobj, jobject nativePointer) {
	JnmDecodingPlayer *jnmDecPlayer = (*env)->GetDirectBufferAddress(env,
			nativePointer);
	jnm_decodingplayer_stop(jnmDecPlayer);
}

JNIEXPORT jboolean JNICALL Java_de_klausj_video_impl_gstreamer_GstMediaDecodingPlayer_seek(
		JNIEnv *env, jobject jobj, jobject nativePointer,jlong reqPos) {
	JnmDecodingPlayer *jnmDecPlayer = (*env)->GetDirectBufferAddress(env,
			nativePointer);
	return jnm_decodingplayer_seek(jnmDecPlayer,reqPos);
}

JNIEXPORT void JNICALL Java_de_klausj_video_impl_gstreamer_GstMediaDecodingPlayer_frameStep(
		JNIEnv *env, jobject jobj, jobject nativePointer,jlong frames) {
	JnmDecodingPlayer *jnmDecPlayer = (*env)->GetDirectBufferAddress(env,
			nativePointer);
	jnm_decodingplayer_frameStep(jnmDecPlayer,frames);
}

JNIEXPORT void JNICALL Java_de_klausj_video_impl_gstreamer_GstMediaDecodingPlayer_rewind(
		JNIEnv *env, jobject jobj, jobject nativePointer) {
	JnmDecodingPlayer *jnmDecPlayer = (*env)->GetDirectBufferAddress(env,
			nativePointer);
	jnm_decodingplayer_rewind(jnmDecPlayer);
}

JNIEXPORT jlong JNICALL Java_de_klausj_video_impl_gstreamer_GstMediaDecodingPlayer_getDuration(
		JNIEnv *env, jobject jobj, jobject nativePointer) {
	JnmDecodingPlayer *jnmDecPlayer = (*env)->GetDirectBufferAddress(env,
			nativePointer);
	return jnm_decodingplayer_getDuration(jnmDecPlayer);
}

JNIEXPORT jlong JNICALL Java_de_klausj_video_impl_gstreamer_GstMediaDecodingPlayer_getPosition(
		JNIEnv *env, jobject jobj, jobject nativePointer) {
	JnmDecodingPlayer *jnmDecPlayer = (*env)->GetDirectBufferAddress(env,
			nativePointer);
	return jnm_decodingplayer_getPosition(jnmDecPlayer);
}

JNIEXPORT void JNICALL Java_de_klausj_video_impl_gstreamer_GstMediaDecodingPlayer_close(
		JNIEnv *env, jobject jobj, jobject nativePointer) {
	JnmDecodingPlayer *jnmDecPlayer = (*env)->GetDirectBufferAddress(env,
			nativePointer);
	jnm_decodingplayer_close(jnmDecPlayer);

}

JNIEXPORT void JNICALL Java_de_klausj_video_impl_gstreamer_GstMediaDecodingPlayer_clear(
		JNIEnv *env, jobject jobj, jobject nativePointer) {
	JnmDecodingPlayer *mdp = (*env)->GetDirectBufferAddress(env,
			nativePointer);
	if(mdp->imgBufferArr!=NULL){
		(*env)->DeleteWeakGlobalRef(env,mdp->imgBufferArr);
	}
	mdp->videoNativeThreadID=0;
}

JNIEXPORT void JNICALL Java_de_klausj_video_impl_gstreamer_GstMediaDecodingPlayer_release(
		JNIEnv *env, jobject jobj, jobject nativePointer) {

	JnmDecodingPlayer *jnmDecPlayer = (*env)->GetDirectBufferAddress(env,
			nativePointer);
	(*env)->DeleteWeakGlobalRef(env, jnmDecPlayer->imgBufferArr);
	(*env)->DeleteWeakGlobalRef(env, jnmDecPlayer->javaObjRef);
	(*env)->DeleteGlobalRef(env, jnmDecPlayer->jClassRef);

	jnm_decodingplayer_release(jnmDecPlayer);
}

